"""
The english translation

Copyright
---------
Animenosekai
    Original Author, MIT License
"""

from nasse.localization.base import Localization


class EnglishLocalization(Localization):
    # it's the same one as the base
    __native__ = "English"
